@@ -1,9 +1,8 @@
 #' Parse and build URLs
 #'
 #' `url_parse()` parses a URL into its component pieces; `url_build()` does
-#' the reverse, converting a list of pieces into a string URL. See
-#' [rfc3986](https://tools.ietf.org/html/rfc3986) for details of parsing
-#' algorithm.
+#' the reverse, converting a list of pieces into a string URL. See `r rfc(3986)`
+#' for the details of the parsing algorithm.
 #'
 #' @param url For `url_parse()` a string to parse into a URL;
 #'   for `url_build()` a URL to turn back into a string.
@@ -25,7 +24,7 @@
 #' url$query <- list(a = 1, b = 2, c = 3)
 #' url_build(url)
 url_parse <- function(url) {
-  check_string(url, "`url`")
+  check_string(url)
 
   # https://datatracker.ietf.org/doc/html/rfc3986#appendix-B
   pieces <- parse_match(url, "^(([^:/?#]+):)?(//([^/?#]*))?([^?#]*)(\\?([^#]*))?(#(.*))?")
@@ -44,10 +43,9 @@ url_parse <- function(url) {
 
   userinfo <- pieces[[2]]
   if (!is.null(userinfo)) {
-    if (grepl(":", userinfo)) {
-      userinfo <- parse_in_half(userinfo, ":")
-    } else {
-      userinfo <- list(userinfo, NULL)
+    userinfo <- parse_in_half(userinfo, ":")
+    if (userinfo$right == "") {
+      userinfo$right <- NULL
     }
   }
   hostname <- pieces[[3]]
@@ -57,8 +55,8 @@ url_parse <- function(url) {
     list(
       scheme = scheme,
       hostname = hostname,
-      username = userinfo[[1]],
-      password = userinfo[[2]],
+      username = userinfo$left,
+      password = userinfo$right,
       port = port,
       path = path,
       query = query,
@@ -68,9 +66,9 @@ url_parse <- function(url) {
   )
 }
 
-url_modify <- function(url, ...) {
+url_modify <- function(url, ..., error_call = caller_env()) {
   url <- url_parse(url)
-  url <- modify_list(url, ...)
+  url <- modify_list(url, ..., error_call = error_call)
   url_build(url)
 }
 
@@ -100,7 +98,9 @@ print.httr2_url <- function(x, ...) {
   if (!is.null(x$query)) {
     cli::cli_li("{.field query}: ")
     id <- cli::cli_ul()
-    cli::cli_li(paste0("  {.field ", names(x$query), "}: ", x$query))
+    # escape curly brackets for cli by replacing single with double brackets
+    query_vals <- gsub("\\{", "{{", gsub("\\}", "}}", x$query))
+    cli::cli_li(paste0("  {.field ", names(x$query), "}: ", query_vals))
     cli::cli_end(id)
   }
   if (!is.null(x$fragment)) {
@@ -121,7 +121,7 @@ url_build <- function(url) {
   if (is.null(url$username) && is.null(url$password)) {
     user_pass <- NULL
   } else if (is.null(url$username) && !is.null(url$password)) {
-    abort("Cannot set url password without username")
+    cli::cli_abort("Cannot set url {.arg password} without {.arg username}.")
   } else if (!is.null(url$username) && is.null(url$password)) {
     user_pass <- paste0(url$username, "@")
   } else {
@@ -137,6 +137,10 @@ url_build <- function(url) {
     authority <- NULL
   }
 
+  if (!is.null(url$path) && !startsWith(url$path, "/")) {
+    url$path <- paste0("/", url$path)
+  }
+
   prefix <- function(prefix, x) if (!is.null(x)) paste0(prefix, x)
   paste0(
     url$scheme, if (!is.null(url$scheme)) ":",
@@ -160,9 +164,9 @@ query_parse <- function(x) {
   out
 }
 
-query_build <- function(x) {
+query_build <- function(x, error_call = caller_env()) {
   if (!is_list(x) || (!is_named(x) && length(x) > 0)) {
-    abort("Query must be a named list")
+    cli::cli_abort("Query must be a named list.", call = error_call)
   }
 
   x <- compact(x)
@@ -172,17 +176,29 @@ query_build <- function(x) {
 
   bad_val <- lengths(x) != 1 | !map_lgl(x, is_atomic)
   if (any(bad_val)) {
-    abort(c(
-      "Query parameters must be length 1 atomic vectors.",
-      paste0("Problems: ", paste0(names(x)[bad_val], collapse =", "))
-    ))
+    cli::cli_abort(
+      c(
+        "Query parameters must be length 1 atomic vectors.",
+        "*" = "Problems: {.str {names(x)[bad_val]}}."
+      ),
+      call = error_call
+    )
   }
 
-  is_double <- map_lgl(x, is.double)
-  x[is_double] <- map_chr(x[is_double], format, scientific = FALSE)
-
   names <- curl::curl_escape(names(x))
-  values <- map_chr(x, curl::curl_escape)
+  values <- map_chr(x, format_query_param, error_call = error_call)
 
   paste0(names, "=", values, collapse = "&")
 }
+
+
+format_query_param <- function(x, error_call = caller_env()) {
+  if (inherits(x, "AsIs")) {
+    x <- unclass(x)
+    check_string(x, call = error_call, arg = I("Escaped query value"))
+    return(x)
+  }
+
+  x <- format(x, scientific = FALSE, trim = TRUE, justify = "none")
+  curl::curl_escape(x)
+}
